/* start of file Laguer.cpp ----------------------------------------------- */

/* <summary>
	SUBROUTINE Laguer - Computes one root on the complex polynomial
	a = a[0], a[1],..., a[n-1]
   </summary>
*/

#include "stdafx.h"
#include "Laguer.h"

using namespace std;

void Laguer(complex<double> a[], int n, int *its, complex<double> *x)
/* x = a root of a, x input is the initial guess for a root
*/
/* The following routine implements the Laguerre method to find one
   root of a given polynomial of degree m = n-1, whose coefficients can
   be complex. The coefficient a[m] is the constant term, while a[0] is
   the coefficient of the highest power of x. The routine implements a
   simplified version of an elegant stopping criterion due to Adams
   [5], which neatly balances the desire to achieve full machine
   accuracy, on the one hand, with the danger of iterating forever in
   the presence of roundoff error, on the other.

   This routine improves x by Laguerre's method until it converges,
   within the achievable roundoff limit, to a root of the given
   polynomial. The number of iterations taken is returned as its.
   Parameters: EPSS is the estimated fractional roundoff error. We try
   to break any limit cycles with mR different fractional values, once
   every mT steps.
*/
{
	double EPSS = 4.0E-16;	/* estimated fractional roundoff error */
	int m = n - 1;	/* n = size of polynomial, m = degree of polynomial */
	int mR = 8;
	int mT = 10;
	int iter, itToGo = 0, j;
	double abb;
	double abx;
	double abp;
	double abm;
	complex<double> cZero = complex<double>(0.0, 0.0);
	double err;
	complex<double> mc = complex<double>((double)m, 0.0);
	complex<double> mmc = complex<double>((double)(m - 1), 0.0);
	complex<double> dx;
	complex<double> xx;
	complex<double> x1;
	complex<double> b;
	complex<double> d;
	complex<double> f;
	complex<double> g;
	complex<double> h;
	complex<double> sq;
	complex<double> gp;
	complex<double> gm;
	complex<double> g2;
	double frac[8];		/* Fractions used to break limit cycle. */
	frac[0] = 0.50;
	frac[1] = 0.25;
	frac[2] = 0.75;
	frac[3] = 0.13;
	frac[4] = 0.38;
	frac[5] = 0.62;
	frac[6] = 0.88;
	frac[7] = 1.00;
	xx = *x;				/* initial guess */
	double tf;
	/* Loop over iterations up to allowed max */
	for (iter = 1; ; iter++)
	{
		*its = iter;
		b = a[0];
		err = abs(b);
		d = 0;
		f = 0;
		abx = abs(xx);
		/* Efficient computation of the polynomial and its first */
		for (j = 1; j < n; j++)
		{
			f = xx * f + d;		/* two derivatives. f stores P''/2. */
			d = xx * d + b;
			b = xx * b + a[j];
			abb = abs(b);
			err = abb + abx * err;
		}
		err *= EPSS;	/* Estimate of roundoff error in evaluating polynomial */
		if (abb <= err && itToGo == 0)
			itToGo = 2 * iter;
		if (b == cZero || (abb <= err && iter >= itToGo))		/* We are on the root */
		{
			int qq = 0;		/* for testing */
			qq++;
			break;
		}
		else	/* The generic case: use Laguerre's formula */
		{
			g = d / b;
			g2 = g * g;
			h = g2 - (f + f) / b;
			/* (m - 1) * (m * h - g2) */
			sq = mmc * (mc * h - g2);
			sq = sqrt(sq);
			gp = g + sq;
			gm = g - sq;
			abp = abs(gp);
			abm = abs(gm);
			if(abp < abm)
				gp = gm;
			tf = abp;
			if (abp < abm)	/* tf = Max(abp, abm) */
				tf = abm;
			if (tf > 0.0)
				dx = mc / gp;
			else
			{
				tf = log(1.0 + abx);
				/* dx = tf + iter * I */
				dx = complex<double>(tf, (double)iter);
				dx = exp(dx);
			}
		}
		x1 = xx - dx;
		if (xx == x1)
			break;	/* Converged */
		j = iter % mT;
		if (j != 0)
			xx = x1;
		else	/* Every so often we take a fractional step, to break any */
				/* limit cycle (itself a rare occurrence) */
		{
			j = (-1 + iter / mT) % mR;
			xx -= dx * frac[j];
		}
	}
	*x = xx;
} /* Laguer */

/* end of file Laguer.cpp ----------------------------------------------- */


